/* -*- c-basic-offset: 3; mode: c++ -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 * 
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

/**
 * Some functions for computing the overlapping matrix.
 * Two algorithms are used:
 * 1/ best matching (Ozdemir et al. 2009)
 * 2/ bipartite matching (implemented with the Hungarian algorithm)
 */

#ifndef _MATCHING_H
#define _MATCHING_H

#include <limits>
#include <pandore.h>
using namespace pandore;

struct PandoreException : public std::exception {
      PandoreException( std::string message ) : _message(message) {}
      ~PandoreException() throw () { }
      
      const char* what() const throw() { return _message.c_str(); }
   private:
      std::string _message;
};

const bool BEST_MATCHING_ID = 0;

class MatchingGraph {
   private:
      Graph2d *_graph;
      int _numberOfSegments;
      int _numberOfRegions;
      bool _ownData;
   
   public:
      MatchingGraph( int numberOfSegments, int numberOfRegions ): _numberOfSegments(numberOfSegments), _numberOfRegions(numberOfRegions), _ownData(false) { }
      
      void computeMatching( const Reg2d &segmentation, const Reg2d &reference, int matchingAlgorithmId, float threshold ) {
	 _ownData = true;
	 _graph = new Graph2d(_numberOfSegments + _numberOfRegions, _numberOfSegments, _numberOfRegions, false);
	 if (reference.Labels() == 1) {
	    computeOverlappingAreaMatrix(segmentation, reference,* _graph);
	    thresholdOverlappingAreaMatrix(*_graph, threshold);
//	    display(*_graph);
	 } else {
	    Graph2d *overlappingAreaMatrix = new Graph2d(_numberOfSegments + _numberOfRegions, _numberOfSegments, _numberOfRegions, false);
	    computeOverlappingAreaMatrix(segmentation, reference, *overlappingAreaMatrix);
	    thresholdOverlappingAreaMatrix(*overlappingAreaMatrix, threshold);

	    if (matchingAlgorithmId == BEST_MATCHING_ID) {
	       bestMatching(*overlappingAreaMatrix);
	    } else {
	       bipartiteMatching(*overlappingAreaMatrix);
	    }
	    delete overlappingAreaMatrix;
	 }
      }
   
      ~MatchingGraph() {
	 if (_ownData) {
	    delete _graph;
	 }
      }
   
      void setMatchingGraph( Graph2d *graph ) {
	 _graph = graph;
      }
   
      Graph2d *getGraph() {
	 return _graph; 
      }
   
      long getRegionArea( int region ) const {
	 const int ri = region + _numberOfSegments;
	 if (!(*_graph)[ri]) {
	    return 0;
	 }
	 return(long)((*_graph)[ri]->value);
      }
   
      long getSegmentArea( int segment ) const {
	 const int si = segment + 0;
	 if (!(*_graph)[si]) {
	    return 0;
	 }
	 return (long)((*_graph)[si]->value);
      }
   
      void setRegionArea( int r, long area ) const {
	 const int ri = r + _numberOfSegments;
	 (*_graph)[ri]->value = (Double)area;
      }

      void setSegmentArea( int s, long area ) const {
	 const int si = s + 0;
	 (*_graph)[si]->value = (Double)area;
      }

      bool isRegionMatched( int r ) const {
	 int ri = r + _numberOfSegments;
	 return (*_graph)[ri] && (*_graph)[ri]->Neighbours();
      }
   
      bool isSegmentMatched( int s ) const {
	 int si = s + 0;
	 return (*_graph)[si] && (*_graph)[si]->Neighbours();
      }

      bool isInverseRegionMatched( int r ) const {
	 int ri = r + _numberOfSegments;
	 return (*_graph)[ri] && (*_graph)[ri]->Neighbours();
      }

      bool isInverseSegmentMatched( int s ) const {
	 int si = s + 0;
	 return (*_graph)[si] && (*_graph)[si]->Neighbours();
      }
   
      GEdge * getMatchedSegmentsList( int r ) const {
	 int ri = r + _numberOfSegments; 
	 return (*_graph)[ri]->Neighbours();
      }

      GEdge * getMatchedRegionsList( int s ) const {
	 int si = s + 0;
	 return (*_graph)[si]->Neighbours();
      }

      GEdge * getInverseMatchedSegmentsList( int r ) const {
	 int ri = r + _numberOfSegments; 
	 return (*_graph)[ri]->Neighbours();
      }

      GEdge * getInverseMatchedRegionsList( int s ) const {
	 int si = s + 0;
	 return (*_graph)[si]->Neighbours();
      }

      bool isRegionMatchedBy( int r, int s ) const {
	 int ri = r + _numberOfSegments;
	 int si = s + 0;
	 return (*_graph)[ri]->Search(si) != 0;
      }
   
      bool isSegmentMatchedBy( int s, int r ) const {
	 int ri = r + _numberOfSegments;
	 int si = s + 0;
	 return (*_graph)[si]->Search(ri) != 0;
      }

   private:
   
      inline int segmentIndexToNodeIndex( int segment ) const {
	 return segment + 0;
      }

      inline int regionIndexToNodeIndex( int region ) const {
	 return region + _numberOfSegments;
      }

      inline int nodeIndexToSegmentIndex( int node ) const {
	 return node - 0;
      }

      inline int nodeIndexToRegionIndex( int node ) const {
	 return node - _numberOfSegments;
      }

      template < typename T >
      void computeOverlappingAreaMatrix( const T &segmentation, const T &reference, Graph2d &overlappingAreaMatrix ) const {
	 Ulong *psegmentation = segmentation.Vector();
	 const Ulong *pend = psegmentation + segmentation.VectorSize();
	 Ulong *preference = reference.Vector();

	 for (; psegmentation < pend; psegmentation++, preference++) {
	    int s = *psegmentation;
	    int si = segmentIndexToNodeIndex(s);
	    int r = *preference;
	    int ri = regionIndexToNodeIndex(r);

	    if (!(*_graph)[si]) {
	       _graph->Add(si, si);
	       (*_graph)[si]->value = 1.0;
	    } else {
	       (*_graph)[si]->value++;
	    }
	    if (!(*_graph)[ri]) {
	       _graph->Add(ri, ri);
	       (*_graph)[ri]->value = 1.0;
	    } else {
	       (*_graph)[ri]->value++;
	    }

	    // Update the overlapping areas.
	    if (r > 0 && s > 0) {
	       if (!overlappingAreaMatrix[si]) {
		  overlappingAreaMatrix.Add(si, si);
	       }
	       if (!overlappingAreaMatrix[ri]) {
		  overlappingAreaMatrix.Add(ri, ri);
	       }
	       overlappingAreaMatrix.Link(si, ri, 1.0, true);
	    }
	 }
      }
      
      void thresholdOverlappingAreaMatrix(Graph2d &overlappingAreaMatrix, float threshold) {
	 for (int s = 1; s < _numberOfSegments ; s++) {
	    int si = segmentIndexToNodeIndex(s);
	    if (overlappingAreaMatrix[si] && overlappingAreaMatrix[si]->Neighbours()) {	       
	       for (GEdge *ptr = overlappingAreaMatrix[si]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		  int ri = ptr->Node();
		  int r = nodeIndexToRegionIndex(ri);
		  if ( (ptr->weight / getRegionArea(r)) < threshold &&
		       (ptr->weight / getSegmentArea(s)) < threshold) {
		     overlappingAreaMatrix.Unlink(si, ri);
		  }
	       }
	    }
	 }
      }

      /**
       * Checks whether the edge between s and r is satisfied or not.
       * Assumption: le link between s and r exists.
       *
       * An edge is satisfied if it is :
       *    - 0
       *    - or 4-min(multiplicity(s), 2)-min(multiplicity(r), 2) >= 1,
       * where multiplicity(s) is the number of edges connected to s, and
       * where multiplicity(r) is the number of edges connected to r.
       * This means that r cannot be associated to s if either r or s is already
       * connected to several others r or s.
       */
      bool isSatisfied( const Graph2d &tab, const int s, const int r ) const {
	 int smultiplicity = 0;
	 for (GEdge * ptr = tab[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
	    if (ptr->weight > 0) {
	       smultiplicity++;
	    }
	 }
	 int rmultiplicity = 0;
	 for (GEdge * ptr = tab[r]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
	    if (ptr->Node() == s) {
	    }
	    if (ptr->weight > 0) {
	       rmultiplicity++;
	    }
	 }
	 return smultiplicity <= 1 || rmultiplicity <= 1;
      }
      
      /**
       * Tests whether all the nodes are consistent.
       * @return 0 if at least one node is unconsistent, the total area otherwise.
       */
      int isGoalState( const Graph2d &tab ) const {
	 int max = 0;
	 for (int s = 1; s < tab.Height(); s++) {
	    if (tab[s]) {
	       for (GEdge *ptr = tab[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		  int r = ptr->Node();
		  if (!isSatisfied(tab, s, r)) {
		     return 0;
		  }
		  max += (int)ptr->weight;
	       }
	    }
	 }
	 return max;
      }
      
      void satisfyEdgeFromSPerspective( Graph2d & tab, Img1dsl& segments, const int s, const int r ) const {
	 const int soffset = 0;
	 for (GEdge * ptr = tab[r]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
	    int si = ptr->Node() - soffset;
	    segments(si) = (int)ptr->weight;
	 }
	 for (int si = 1; si < tab.Height(); si++) {
	    if (si != s - soffset && segments(si) > 0) {
	       tab.Unlink(si + soffset, r);
	    }
	 }
      }
   
      void satisfyEdgeFromRPerspective( Graph2d & tab, Img1dsl& regions, const int s, const int r ) const {
	 const int roffset = tab.Height();
	 for (GEdge * ptr = tab[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
	    int ri= ptr->Node() - roffset;
	    regions(ri)=(Ulong)ptr->weight;
	 }
	 for (int ri = 1; ri < tab.Width(); ri++) {
	    if (ri != r - roffset && regions(ri) > 0) {
	       tab.Unlink(s, ri + roffset);
	    }
	 }
      }

      void restoreEdgeFromSPerspective( Graph2d & tab, Img1dsl& segments, const int r ) const {
	 const int soffset = 0;
	 for (int si = 1; si < tab.Height(); si++) {
	    if (segments(si) > 0) {
	       tab.Link(si + soffset, r, (double)segments(si));
	    }
	 }
      }
 
      void restoreEdgeFromRPerspective( Graph2d & tab, const Img1dsl& regions, const int s ) const {
	 const int roffset = tab.Height();
	 for (int ri = 1; ri < tab.Width(); ri++) {
	    if (regions(ri) > 0) {
	       tab.Link(s, ri + roffset, (double)regions(ri));
	    }
	 }
      }
   
      double computeWeightRi( const Graph2d & oam, const int, const int r ) const {
	 double w = 0;
	 for (GEdge *ptr = oam[r]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
	    w += ptr->weight;
	 } 
	 return w;
      }
   
      double computeWeightSi( const Graph2d & oam, const int s, const int ) const {
	 double w = 0;
	 for (GEdge *ptr = oam[s]->Neighbours(); ptr != NULL; ptr=ptr->Next()) {
	    w += ptr->weight;
	 } 
	 return w;
      }

      double computeArea( const Graph2d &tab ) const {
	 double area = 0;
	 for (int s = 1; s < tab.Height(); s++) {
	    if (tab[s]) {
	       for (GEdge *ptr = tab[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		  area += ptr->weight;
	       }
	    }
	 }
	 return area;
      }

      void display( Graph2d& g ) const {
	 puts("****  GRAPH ************");
	 for (int si = 0; si < g.Height(); si++) {
	    if (!g[si]) { continue; }
	    bool edge = false;
	    for (GEdge *ptr = g[si]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
	       edge = true;
	       printf ("s=%d -> r=%d = (area = %.0f)\t", si, ptr->Node() - g.Height(), ptr->weight);
	    }
	    if (edge) {
	       printf ("\n");
	    }
	 }
	 puts("\n");
      }
   
      /**
       * Finds the edge with the maximum area.
       * The resulted edge is between smax and rmax.
       */
      double findMaximumEdge( Graph2d &oam, int &smax, int &rmax ) const {
	 double max = 0;

	 for (int s = 1; s < oam.Height(); s++) {
	    if (oam[s]) {
	       for (GEdge *ptr = oam[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		  int r = ptr->Node();
		  if (!isSatisfied(oam, s, r) && ptr->weight > max) {
		     max = ptr->weight;
		     smax = s;
		     rmax = r;
		  }
	       }
	    }
	 }
	 return max;
      }

      double optimize( Graph2d &tab, Graph2d &tabmax, double max ) const {
	 int smax = 0;
	 int rmax = 0;

	 double weight = findMaximumEdge(tab, smax, rmax);

	 Img1dsl segments(tab.Height());
	 Img1dsl regions(tab.Width()); 

	 if (smax > 0) {
	    double max1;
	    double w1 = computeWeightRi(tab, smax, rmax);
	    double w2 = computeWeightSi(tab, smax, rmax);

	    if (w1 > w2) {
	       // 1. Consider rmax as the main region that has a multiplicity>1.
	       // Remove the edges of the si that are connected to rmax
	       // other than to the edge rmax-si.
	       regions = 0;
	       satisfyEdgeFromRPerspective(tab, regions, smax, rmax);
	       max1 = optimize(tab, tabmax, max);
	       if (max1 > max) {
		  max = max1;
	       }
	       restoreEdgeFromRPerspective(tab, regions, smax);

	       // 2. Consider si as the main region that has a multiplicity>1.
	       // Remove the edges of the ri that are connected to smax
	       // other than to the edge ri-smax.
	       segments = 0;
	       satisfyEdgeFromSPerspective(tab, segments, smax, rmax);
	       if (computeArea(tab) > max) {
		  max1 = optimize(tab, tabmax, max);
		  if (max1 > max) {
		     max = max1;
		  }
	       }
	       restoreEdgeFromSPerspective(tab, segments, rmax);
	    } else {
	       // 1. Consider rmax as the main region that has a multiplicity>1.
	       // Remove the edges of the ri that are connected to smax
	       // other than to the edge ri-smax.
	       segments = 0;
	       satisfyEdgeFromSPerspective(tab, segments, smax, rmax);
	       max1=optimize(tab, tabmax, max);
	       if (max1 > max) {
		  max = max1;
	       }
	       restoreEdgeFromSPerspective(tab, segments, rmax);
	    
	       // 2. Consider si as the main region that has a multiplicity>1.
	       // Remove the edges of the si that are connected to rmax
	       // other than to the edge si-rmax.
	       regions = 0;
	       satisfyEdgeFromRPerspective(tab, regions, smax, rmax);
	       if (computeArea(tab) > max) {
		  max1 = optimize(tab, tabmax, max);
		  if (max1 > max) {
		     max = max1;
		  }
	       }
	       restoreEdgeFromRPerspective(tab, regions, smax);
	    }

	    // 3. Remove the edge.
	    tab.Unlink(smax, rmax);

	    // If the new state have an area < max then stop the optimization
	    if (computeArea(tab) > max) {
	       max1 = optimize(tab, tabmax, max);
	       if (max1 > max) {
		  max = max1;
	       }
	    }
	    tab.Link(smax, rmax, weight);
	    return max;
	 } else {
	    // When all the edges are satisfied.
	    int max1 = isGoalState(tab);
	    if (max1 > max) {
	       tabmax = tab;
	       max = max1;
	    }
	    return max1;
	 }
      }

      int getClique( Graph2d &grs, Graph2d &grd, int r ) {
	 int x = 0;
	 if (!grs[r]) { 
	    return 0;
	 }
	 GEdge *ptr = grs[r]->Neighbours();
	 if (!grd[r]) {
	    grd.Add(r, r);
	 }
	 while (ptr != NULL) {
	    int s = ptr->Node();
	    if (!grd[s]) {
	       grd.Add(s, s);
	    }
	    grd.Link(r, s, ptr->weight);
	    grs.Unlink(r, s);
	    x += 1 + getClique(grs, grd, s); 
	    ptr = grs[r]->Neighbours();
	 }
	 return x;
      }

      /**
       * Computes the best matching between the segmentation
       * and the reference. The algorithm is based on
       * constraint satisfaction.
       */
      void bestMatching( Graph2d &overlappingAreaMatrix ) {
	 Graph2d subGraph;
	 Graph2d bestMatchingGraph;

	 for (int s = 1; s < _numberOfSegments ; s++) {
	    if (overlappingAreaMatrix[s] && overlappingAreaMatrix[s]->Neighbours()) {
	       subGraph.New(overlappingAreaMatrix.Props());
	       getClique(overlappingAreaMatrix, subGraph, s);
	       bestMatchingGraph.New(overlappingAreaMatrix.Props());
	       double max = optimize(subGraph, bestMatchingGraph, 0.0);
	       if (max > 0) {
		  for (int si = 1; si < _numberOfSegments; si++) {
		     if (bestMatchingGraph[si]) {
			for (GEdge *ptr = bestMatchingGraph[si]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
			   if (ptr->weight > 0) {
			      int ri = ptr->Node();

			      if (!(*_graph)[si]) {
				 _graph->Add(si, si);
			      }
			      if (!(*_graph)[ri]) {
				 _graph->Add(ri, ri);
			      }
			      _graph->Link(si, ri, ptr->weight);
			   }
			}
		     }
		  }
	       }
	    }
	 }
      }

      // BIPARTITE 
      Img2dsl _costMatrix;
      Img2duc _maskMatrix;  // Used to star and prime zeros of the cost matrix.
      int ** _path;
      int *_rowCover;
      int *_colCover;
      int path_row_0;
      int path_col_0;
      int step;  // the current step in the processing (state machine).
      bool _invert; // true if the width < height;

      void computeBipartiteGraph( const Graph2d &overlappingAreaMatrix ) {
	 bool done = false;
	 step = 1;
	 while (!done) {
	    switch (step) {
	       case 1:
		  step = step1();
		  break;
	       case 2:
		  step = step2();
		  break;
	       case 3:
		  step = step3();
		  break;
	       case 4:
		  step = step4();
		  break;
	       case 5:
		  step = step5();
		  break;
	       case 6:
		  step = step6();
		  break;
	       case 7:
		  done = true;
		  break;
	    }
	 }
	 if (_invert) {
	    for (int si = 0; si < _numberOfSegments; si++) {
	       if (overlappingAreaMatrix[si]) {
		  for (GEdge *ptr = overlappingAreaMatrix[si]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		     int ri = ptr->Node();
		     int s = si - 0;
		     int r = nodeIndexToRegionIndex(ri);

		     if (_maskMatrix(s, r) == 1) {
			if (!(*_graph)[si]) { 
			   _graph->Add(si, si);
			}
			if (!(*_graph)[ri]) {
			   _graph->Add(ri, ri);
			}
			_graph->Link(si, ri, ptr->weight);
		     }
		  }
	       }
	    }
	 } else {
	    for (int si = 0; si < _numberOfSegments; si++) {
	       if (overlappingAreaMatrix[si]) {
		  for (GEdge *ptr = overlappingAreaMatrix[si]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		     int ri = ptr->Node();
		     int s = si - 0;
		     int r = nodeIndexToRegionIndex(ri);
		     if (_maskMatrix(r, s) == 1) {
			if (!(*_graph)[si]) {
			   _graph->Add(si, si);
			}
			if (!(*_graph)[ri]) {
			   _graph->Add(ri, ri);
			}
			_graph->Link(si, ri, ptr->weight);
		     }
		  }
	       }
	    }
	 }
      }

      /**
       * For each row of the cost matrix,
       * finds the smallest element and subtracts
       * it from every element in its row. 
       * When finished, Go to Step 2.
       */
      int step1( ) {
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    int min_in_row = _costMatrix(r, 0);
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_costMatrix(r, c) < min_in_row) {
		  min_in_row = _costMatrix(r, c);
	       }
	    }
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       _costMatrix(r, c) -= min_in_row;
	    }
	 }

	 return 2;
      }

      /**
       * Finds a zero (Z) in the resulting matrix. If there is no starred 
       * zero in its row or column, stars Z. Repeats for each element in the 
       * matrix. Go to Step 3.
       */
      int step2( ) {
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_costMatrix(r, c) == 0 && _rowCover[r] == 0 && _colCover[c] == 0) {
		  _maskMatrix(r, c) = 1;
		  _rowCover[r] = 1;
		  _colCover[c] = 1;
	       }
	    }
	 }
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    _rowCover[r] = 0;
	 }
	 for (int c = 0; c < _costMatrix.Width(); c++) {
	    _colCover[c] = 0;
	 }
	 return 3;
      }
   
      /**
       * Covers each column containing a starred zero. If K columns are covered, 
       * the starred zeros describe a complete set of unique assignments. In this 
       * case, Go to DONE, otherwise, Go to Step 4.
       */
      int step3( ) {
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_maskMatrix(r, c) == 1) {
		  _colCover[c] = 1;
	       }
	    }
	 }
	 int colcount = 0;
	 for (int c = 0; c < _costMatrix.Width(); c++) {
	    if (_colCover[c] == 1) {
	       colcount += 1;
	    }
	 }
	 if (colcount >= _costMatrix.Width() || colcount >= _costMatrix.Height()) {
	    return 7;
	 }
	 return 4;
      }
   
      Point2d find_a_zero() {
	 Point2d p(-1, -1);

	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_costMatrix(r, c) == 0 && _rowCover[r] == 0 && _colCover[c] == 0) {
		  p.y = r;
		  p.x = c;
		  return p;
	       }
	    }
	 }
	 return p;
      }

      int find_star_in_row( int row ) {
	 for (int c = 0; c < _costMatrix.Width(); c++) {
	    if (_maskMatrix(row, c) == 1) {
	       return c;
	    }
	 }
	 return -1;
      }

      /**
       * Finds a noncovered zero and primes it. If there is no starred zero 
       * in the row containing this primed zero, Go to Step 5.  Otherwise, 
       * cover this row and uncover the column containing the starred zero. 
       * Continue in this manner until there are no uncovered zeros left. 
       * Save the smallest uncovered value and Go to Step 6.
       */
      int step4( ) {	 
	 while (true) {
	    Point2d p = find_a_zero();
	    int row = p.y;
	    int col = p.x;
	    if (row == -1) {
	       return 6;
	    } else {
	       _maskMatrix(row, col) = 2;
	       int ncol = find_star_in_row(row);
	       if (ncol >= 0) {
		  _rowCover[row] = 1;
		  _colCover[ncol] = 0;
	       } else {
		  path_row_0 = row;
		  path_col_0 = col;
		  return 5;
	       }
	    }
	 }
      }
      
      int find_star_in_col( int c ) {
	 for (int i = 0; i < _costMatrix.Height(); i++) {
	    if (_maskMatrix(i, c) == 1) {
	       return i;
	    }
	 }
	 return -1;
      }
      
      int find_prime_in_row( int r ) {
	 for (int j = 0; j < _costMatrix.Width(); j++) {
	    if (_maskMatrix(r, j) == 2) {
	       return j;
	    }
	 }
	 return 0;
      }

      void augment_path( int path_count ) {
	 for (int p = 0; p < path_count; p++) {
	    if (_maskMatrix(_path[p][0], _path[p][1]) == 1) {
	       _maskMatrix(_path[p][0], _path[p][1]) = 0;
	    } else {
	       _maskMatrix(_path[p][0], _path[p][1]) = 1;
	    }
	 }
      }
      
      void clear_covers() {
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    _rowCover[r] = 0;
	 }
	 for (int c = 0; c < _costMatrix.Width(); c++) {
	    _colCover[c] = 0;
	 }
      }
      
      void erase_primes() {
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_maskMatrix(r, c) == 2) {
		  _maskMatrix(r, c) = 0;
	       }
	    }
	 }
      }
      
      /**
       * Constructs a series of alternating primed and starred zeros as follows.
       * Let Z0 represents the uncovered primed zero found in Step 4.
       * Let Z1 denotes the starred zero in the column of Z0 (if any).
       * Let Z2 denotes the primed zero in the row of Z1 (there will always be one).
       * Continue until the series terminates at a primed zero that has no
       * starred zero in its column.
       * Unstars each starred zero of the series, star each primed zero of the
       * series, erases all primes and uncover every line in the matrix.
       * Returns to Step 3.
       */
      int step5( ) {
	 int path_count = 1;
	 _path[path_count - 1][0] = path_row_0;
	 _path[path_count - 1][1] = path_col_0;
	 while (true) {
	    int r = find_star_in_col(_path[path_count - 1][1]);
	    if (r<0) break;
	    path_count += 1;
	    _path[path_count - 1][0] = r;
	    _path[path_count - 1][1] = _path[path_count - 2][1];
	 
	    int c= find_prime_in_row(_path[path_count - 1][0]);
	    path_count += 1;
	    _path[path_count - 1][0] = _path[path_count - 2][0];
	    _path[path_count - 1][1] = c;
	 }
	 augment_path(path_count);
	 clear_covers();
	 erase_primes();
	 return 3;
      }

      int find_smallest() {
#undef max
	 int minval= std::numeric_limits<int>::max();
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_rowCover[r] == 0 && _colCover[c] == 0) {
		  if (minval > _costMatrix(r, c)) {
		     minval = _costMatrix(r, c);
		  }
	       }
	    }
	 }
	 return minval;
      }

      /**
       * Adds the value found in Step 4 to every element of
       * each covered row, and subtracts it from every element of
       * each uncovered column.
       * Return to Step 4 without altering any stars, primes, or covered lines.
       */
      int step6(  ) {
	 int minval = find_smallest();
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    for (int c = 0; c < _costMatrix.Width(); c++) {
	       if (_rowCover[r] == 1) {
		  _costMatrix(r, c) += minval;
	       }
	       if (_colCover[c] == 0) {
		  _costMatrix(r, c) -= minval;
	       }
	    }
	 }
	 return 4;
      }

      void bipartiteMatching( const Graph2d &overlappingAreaMatrix ) { 
	 if (_numberOfSegments < _numberOfRegions) {
	    _costMatrix.New(_numberOfSegments, _numberOfRegions);
	    _costMatrix = 0;
	    _invert = true;

	    for (int s = 0; s < _numberOfSegments; s++) {
	       if (overlappingAreaMatrix[s]) {
		  for (GEdge * ptr = overlappingAreaMatrix[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		     int r = nodeIndexToRegionIndex(ptr->Node());
		     _costMatrix(s, r) = -1 * (Long)ptr->weight;
		  }
	       }
	    }
	 } else {
	    _costMatrix.New(_numberOfRegions, _numberOfSegments);
	    _costMatrix = 0;
	    _invert = false;

	    for (int s = 0; s < _numberOfSegments; s++) {
	       if (overlappingAreaMatrix[s]) {
		  for (GEdge * ptr = overlappingAreaMatrix[s]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		     int r = nodeIndexToRegionIndex(ptr->Node());
		     _costMatrix(r, s) = -1 * (Long)ptr->weight;
		  }
	       }
	    }
	 }
	 _maskMatrix.New(_costMatrix.Props());
	 _maskMatrix = 0;
	 _rowCover = new int[_costMatrix.Height()];
	 _colCover = new int[_costMatrix.Width()];
	 
	 for (int r = 0; r < _costMatrix.Height(); r++) {
	    _rowCover[r] = 0;
	 }
	 for (int c = 0; c < _costMatrix.Width(); c++) {
	    _colCover[c] = 0;
	 }

	 _path = new int*[_costMatrix.Height() + _costMatrix.Width()]; 
	 for (int i = 0; i < _costMatrix.Height() + _costMatrix.Width(); i++) {
	    _path[i] = new int[2];
	 }
	 
	 computeBipartiteGraph(overlappingAreaMatrix);

	 delete[] _rowCover;
	 delete[] _colCover;

	 for (int i = 0; i < _costMatrix.Height() + _costMatrix.Width(); i++) {
	    delete[] _path[i];
	 }
	 delete[] _path;
      }

      void noMatching( Graph2d &overlappingAreaMatrix ) { 
	 Graph2d subGraph;
	 Graph2d bestMatchingGraph;

	 for (int si = 1; si < _numberOfSegments; si++) {
	    if (overlappingAreaMatrix[si]) {
	       for (GEdge *ptr = overlappingAreaMatrix[si]->Neighbours(); ptr != NULL; ptr = ptr->Next()) {
		  if (ptr->weight > 0) {
		     int ri = ptr->Node();
		     
		     if (!(*_graph)[si]) {
			_graph->Add(si, si);
		     }
		     if (!(*_graph)[ri]) {
			_graph->Add(ri, ri);
		     }
		     _graph->Link(si, ri, ptr->weight);
		  }
	       }
	    }
	 }
      }
};
#endif
