/* -*- c-basic-offset: 3; mode: c++ -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved.
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 *
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

/**
 * Hole addition / deletion.
 * Recall + Precision
 */

namespace segmentation {
#ifdef MAIN
#undef MAIN
#include "logic/pand.cpp"
#include "segmentation/plabeling.cpp"
#define MAIN
#else
#include "logic/pand.cpp"
#include "segmentation/plabeling.cpp"
#endif
}

#include "evaluation.h"

#define NOTHOLE -2
#define UNVISITED -1

inline void
getHoles( const Reg2d &holeImage, const Reg2d &rgs, long * const neighbors, const int numberOfLabels, const int connexity ) {
   for (int r = 0; r < (int)numberOfLabels + 1; r++) {
      neighbors[r] = UNVISITED;
   }
   Point2d p;
   // A hole is a region that has only one neighbor.
   for (p.y = 0; p.y < holeImage.Height(); p.y++) {
      for (p.x = 0; p.x < holeImage.Width(); p.x++) {
	 Ulong label = holeImage[p];
	 int segment = neighbors[label];
	 if (label > 0 && segment != NOTHOLE) {
	    for (int v = 0; v < connexity; v++) {
	       Point2d offset = (reinterpret_cast<const Point2d*>(vc[connexity]))[v];
	       Point2d q = p + offset;
	       if (holeImage.Hold(q)) {
		  if (holeImage[q] == 0) { // Hole boundary.
		     long region = neighbors[label];
		     if (region == UNVISITED) {
			neighbors[label] = rgs[q];
		     } else if (neighbors[label] != (long)rgs[q]) {
			neighbors[label] = NOTHOLE;
		     }
		  }
	       } else {
		  neighbors[label] = NOTHOLE;
	       }
	    }
	 }
      }
   }
}

inline void
HoleSelection( const Reg2d &rgs, Reg2d& rgd, const int connexity ) {
   Img2duc mismatchedImage(rgs.Props());
   Point2d p;
   
   // Build the mask image with the detected regions only.
   for (p.y = 0; p.y < rgs.Height(); p.y++) {
      for ( p.x = 0; p.x < rgs.Width(); p.x++) {
	 int r = rgs[p];
	 if (r == 0 ) {
	    mismatchedImage[p] = 255;
	 } else {
	    mismatchedImage[p] = 0;
	 }
      }
   }

   Reg2d holeImage(mismatchedImage.Props());
   Ulong numberOfLabels = segmentation::PLabeling(mismatchedImage, holeImage, connexity);
   long *neighbors = new long[numberOfLabels + 1];
   getHoles(holeImage, rgs, neighbors, numberOfLabels, connexity);

   // Count the number of holes.
   for (p.y = 0; p.y < holeImage.Height(); p.y++) {
      for (p.x = 0; p.x < holeImage.Width(); p.x++) {
	 if (neighbors[holeImage[p]] > 0) {
	    rgd[p] = holeImage[p];
	 } else {
	    rgd[p] = 0;
	 }
      }
   }

   delete[] neighbors;
   rgd.Labels(holeImage.Labels());
}

inline int
getNumberOfDetectedHoles( const Reg2d &reference, const Reg2d &segmentation, const int connexity ) {
   Reg2d referenceHoles(reference.Props());
   Reg2d segmentationHoles(segmentation.Props());
   Reg2d commonHoles(reference.Props());

   HoleSelection(reference, referenceHoles, connexity);
   HoleSelection(segmentation, segmentationHoles, connexity);
   int numberOfRegions = segmentation::PAnd(referenceHoles, segmentationHoles, commonHoles);

   return numberOfRegions;
}

/**
 * An added hole is a connected component composed of false negative that are completely inside
 * the related segment.
 */ 
inline DiscrepancyError
computeHoleAdditionError( const Reg2d &reference, const Reg2d &segmentation, const MatchingGraph &graph, const int connexity ) {
   Img2duc mismatchedImage(reference.Props());
   Point2d p;
   
   // Build the mask image with the detected regions only.
   for (p.y = 0; p.y < reference.Height(); p.y++) {
      for ( p.x = 0; p.x < reference.Width(); p.x++) {
	 int r = reference[p];
	 if (r == 0 || !graph.isRegionMatched(r)) {
	    mismatchedImage[p] = 0;
	 } else {
	    int s = segmentation[p];
	    if (s == 0) {
	       mismatchedImage[p] = 255;
	    } else {
	       mismatchedImage[p] = 0;
	    }
	 }
      }
   }
   
   Reg2d holeImage(mismatchedImage.Props());
   
   const int numberOfLabels = segmentation::PLabeling(mismatchedImage, holeImage, connexity);
   long *neighbors = new long[numberOfLabels + 1];
   
   getHoles(holeImage, segmentation, neighbors, numberOfLabels, connexity);
   
   // Count the mean of hole ratio (|hole|/|segment|)
   int holeAdditionRatio = 0;
   for (int r = 0; r < (int)holeImage.Labels() + 1; r++) {
      if (neighbors[r] > 0) {
	 holeAdditionRatio++;
      }
   }
   
   delete[] neighbors;
   
   return DiscrepancyError((float)holeAdditionRatio, 1.0f);
}

/**
 * An deleted hole is a connected component composed of false positive that are completely inside
 * the related region.
 */ 
inline DiscrepancyError
computeHoleDeletionError( const Reg2d &segmentation, const Reg2d &reference, const MatchingGraph &graph, const int connexity ) {
   Img2duc mismatchedImage(segmentation.Props());
   
   // Build the mask image with the detected regions only.
   Point2d p;
   for (p.y = 0; p.y < segmentation.Height(); p.y++) {
      for ( p.x = 0; p.x < segmentation.Width(); p.x++) {
	 int s = segmentation[p];
	 if (s == 0 || !graph.isSegmentMatched(s)) {
	    mismatchedImage[p] = 0;
	 } else {
	    int r = reference[p];
	    if (r == 0) {
	       mismatchedImage[p] = 255;
	    } else {
	       mismatchedImage[p] = 0;
	    }
	 }
      }
   }

   Reg2d holeImage(mismatchedImage.Props());
   const int numberOfLabels = segmentation::PLabeling(mismatchedImage, holeImage, connexity);
   long *neighbors = new long[numberOfLabels + 1];

   getHoles(holeImage, reference, neighbors, numberOfLabels, connexity);

   // Count the mean of hole ratio (|hole|/|segment|)
   int holeAdditionRatio = 0;
   for (int r = 1; r < (int)holeImage.Labels() + 1; r++) {
      if (neighbors[r] > 0) {
	 holeAdditionRatio++;
      }
   }

   delete[] neighbors;

   return DiscrepancyError((float)holeAdditionRatio, 1.0f);
}

inline DiscrepancyIndicator
TopologyPreservationMeasurement( const Reg2d &segmentation, const Reg2d &reference,
				 Graph2d &grs,
				 const int connexity ) {
   if (segmentation.Size() != reference.Size()) {
      throw PandoreException("Error "__FILE__": Incompatible size between the two region maps.");
   }
   if (connexity != 4 && connexity != 8) {
      throw PandoreException("Error "__FILE__": Bad connexity parameter value.");
   }
   const int numberOfSegments = segmentation.Labels() + 1;
   const int numberOfRegions = reference.Labels() + 1;

   DiscrepancyIndicator indicator;
   
   MatchingGraph graph(numberOfSegments, numberOfRegions);
   
   graph.setMatchingGraph(&grs);
   
   // 3. Measures
   
   // 3a. Count the number of correctly detected holes.
   const int holeCount = getNumberOfDetectedHoles(reference, segmentation, connexity);
   
   // 3b. number of added holes.
   indicator.errors[0] = computeHoleAdditionError(reference, segmentation, graph, connexity);
   indicator.errors[0].normalization = holeCount + indicator.errors[0].value;
   
   // 3c. number of deleted holes.
   indicator.errors[1] = computeHoleDeletionError(segmentation, reference, graph, connexity);
   indicator.errors[1].normalization = holeCount + indicator.errors[1].value;
   
#ifndef NDEBUG
   printf("\n* TOPOLOGY PRESERVATION\n");
   printf(" 9/  Hole addition...............: %.0f hole(s) / %.0f detected holes\n", indicator.errors[0].value, indicator.errors[1].normalization);
   printf(" 10/ Hole deletion...............: %.0f hole(s) / %.0f existing holes\n", indicator.errors[1].value, indicator.errors[0].normalization);
#endif

   return indicator;
}
